"""Asynchronous Python client for WLED."""

from functools import lru_cache

from awesomeversion import AwesomeVersion


@lru_cache
def get_awesome_version(version: str) -> AwesomeVersion:
    """Return a cached AwesomeVersion object."""
    return AwesomeVersion(version)



"""Asynchronous Python client for WLED."""

import asyncio
import sys
from typing import Annotated

import typer
from rich.console import Console
from rich.live import Live
from rich.panel import Panel
from rich.table import Table
from zeroconf import ServiceStateChange, Zeroconf
from zeroconf.asyncio import AsyncServiceBrowser, AsyncServiceInfo, AsyncZeroconf

from wled import WLED, WLEDReleases
from wled.exceptions import WLEDConnectionError, WLEDUnsupportedVersionError

from .async_typer import AsyncTyper

cli = AsyncTyper(help="WLED CLI", no_args_is_help=True, add_completion=False)
console = Console()


@cli.error_handler(WLEDConnectionError)
def connection_error_handler(_: WLEDConnectionError) -> None:
    """Handle connection errors."""
    message = """
    Could not connect to the specified WLED device. Please make sure that
    the device is powered on, connected to the network and that you have
    specified the correct IP address or hostname.

    If you are not sure what the IP address or hostname of your WLED device
    is, you can use the scan command to find it:

    wled scan
    """
    panel = Panel(
        message,
        expand=False,
        title="Connection error",
        border_style="red bold",
    )
    console.print(panel)
    sys.exit(1)


@cli.error_handler(WLEDUnsupportedVersionError)
def unsupported_version_error_handler(
    _: WLEDUnsupportedVersionError,
) -> None:
    """Handle unsupported version errors."""
    message = """
    The specified WLED device is running an unsupported version.

    Currently only 0.14.0 and higher is supported
    """
    panel = Panel(
        message,
        expand=False,
        title="Unsupported version",
        border_style="red bold",
    )
    console.print(panel)
    sys.exit(1)


@cli.command("info")
async def command_info(
    host: Annotated[
        str,
        typer.Option(
            help="WLED device IP address or hostname",
            prompt="Host address",
            show_default=False,
        ),
    ],
) -> None:
    """Show the information about the WLED device."""
    with console.status(
        "[cyan]Fetching WLED device information...", spinner="toggle12"
    ):
        async with WLED(host) as led:
            device = await led.update()

    info_table = Table(title="\nWLED device information", show_header=False)
    info_table.add_column("Property", style="cyan bold")
    info_table.add_column("Value", style="green")

    info_table.add_row("Name", device.info.name)
    info_table.add_row("Brand", device.info.brand)
    info_table.add_row("Product", device.info.product)

    info_table.add_section()
    info_table.add_row("IP address", device.info.ip)
    info_table.add_row("MAC address", device.info.mac_address)
    if device.info.wifi:
        info_table.add_row("Wi-Fi BSSID", device.info.wifi.bssid)
        info_table.add_row("Wi-Fi channel", str(device.info.wifi.channel))
        info_table.add_row("Wi-Fi RSSI", f"{device.info.wifi.rssi} dBm")
        info_table.add_row("Wi-Fi signal strength", f"{device.info.wifi.signal}%")

    info_table.add_section()
    info_table.add_row("Version", device.info.version)
    info_table.add_row("Build", str(device.info.build))
    info_table.add_row("Architecture", device.info.architecture)
    info_table.add_row("Arduino version", device.info.arduino_core_version)

    info_table.add_section()
    info_table.add_row("Uptime", f"{int(device.info.uptime.total_seconds())} seconds")
    info_table.add_row("Free heap", f"{device.info.free_heap} bytes")
    info_table.add_row("Total storage", f"{device.info.filesystem.total} bytes")
    info_table.add_row("Used storage", f"{device.info.filesystem.used} bytes")
    info_table.add_row("% Used storage", f"{device.info.filesystem.used_percentage}%")

    info_table.add_section()
    info_table.add_row("Effect count", f"{device.info.effect_count} effects")
    info_table.add_row("Palette count", f"{device.info.palette_count} palettes")

    info_table.add_section()
    info_table.add_row("Sync UDP port", str(device.info.udp_port))
    info_table.add_row(
        "WebSocket",
        "Disabled"
        if device.info.websocket is None
        else f"{device.info.websocket} client(s)",
    )

    info_table.add_section()
    info_table.add_row("Live", "Yes" if device.info.live else "No")
    info_table.add_row("Live IP", device.info.live_ip)
    info_table.add_row("Live mode", device.info.live_mode)

    info_table.add_section()
    info_table.add_row("LED count", f"{device.info.leds.count} LEDs")
    info_table.add_row("LED power", f"{device.info.leds.power} mA")
    info_table.add_row("LED max power", f"{device.info.leds.max_power} mA")

    console.print(info_table)


@cli.command("effects")
async def command_effects(
    host: Annotated[
        str,
        typer.Option(
            help="WLED device IP address or hostname",
            prompt="Host address",
            show_default=False,
        ),
    ],
) -> None:
    """Show the effects on the device."""
    with console.status(
        "[cyan]Fetching WLED device information...", spinner="toggle12"
    ):
        async with WLED(host) as led:
            device = await led.update()

    table = Table(title="\nEffects on this WLED device", show_header=False)
    table.add_column("Effects", style="cyan bold")
    for effect in device.effects.values():
        table.add_row(effect.name)

    console.print(table)


@cli.command("palettes")
async def command_palettes(
    host: Annotated[
        str,
        typer.Option(
            help="WLED device IP address or hostname",
            prompt="Host address",
            show_default=False,
        ),
    ],
) -> None:
    """Show the palettes on the device."""
    with console.status(
        "[cyan]Fetching WLED device information...", spinner="toggle12"
    ):
        async with WLED(host) as led:
            device = await led.update()

    table = Table(title="\nPalettes on this WLED device", show_header=False)
    table.add_column("Palette", style="cyan bold")
    for palettes in device.palettes.values():
        table.add_row(palettes.name)

    console.print(table)


@cli.command("playlists")
async def command_playlists(
    host: Annotated[
        str,
        typer.Option(
            help="WLED device IP address or hostname",
            prompt="Host address",
            show_default=False,
        ),
    ],
) -> None:
    """Show the playlists on the device."""
    with console.status(
        "[cyan]Fetching WLED device information...", spinner="toggle12"
    ):
        async with WLED(host) as led:
            device = await led.update()

    if not device.playlists:
        console.print("🚫[red] This device has no playlists")
        return

    table = Table(title="\nPlaylists stored in the WLED device", show_header=False)
    table.add_column("Playlist", style="cyan bold")
    for playlist in device.playlists.values():
        table.add_row(playlist.name)

    console.print(table)


@cli.command("presets")
async def command_presets(
    host: Annotated[
        str,
        typer.Option(
            help="WLED device IP address or hostname",
            prompt="Host address",
            show_default=False,
        ),
    ],
) -> None:
    """Show the presets on the device."""
    with console.status(
        "[cyan]Fetching WLED device information...", spinner="toggle12"
    ):
        async with WLED(host) as led:
            device = await led.update()

    if not device.presets:
        console.print("🚫[red] This device has no presets")
        return

    table = Table(title="\nPresets stored in the WLED device")
    table.add_column("Preset", style="cyan bold")
    table.add_column("Quick label", style="cyan bold")
    table.add_column("Active", style="green")
    for preset in device.presets.values():
        table.add_row(preset.name, preset.quick_label, "Yes" if preset.on else "No")

    console.print(table)


@cli.command("releases")
async def command_releases() -> None:
    """Show the latest release information of WLED."""
    with console.status(
        "[cyan]Fetching latest release information...", spinner="toggle12"
    ):
        async with WLEDReleases() as releases:
            latest = await releases.releases()

    table = Table(
        title="\n\nFound WLED Releases", header_style="cyan bold", show_lines=True
    )
    table.add_column("Release channel")
    table.add_column("Latest version")
    table.add_column("Release notes")

    table.add_row(
        "Stable",
        latest.stable,
        f"https://github.com/Aircoookie/WLED/releases/v{latest.stable}",
    )
    table.add_row(
        "Beta",
        latest.beta,
        f"https://github.com/Aircoookie/WLED/releases/v{latest.beta}",
    )

    console.print(table)


@cli.command("scan")
async def command_scan() -> None:
    """Scan for WLED devices on the network."""
    zeroconf = AsyncZeroconf()
    background_tasks = set()

    table = Table(
        title="\n\nFound WLED devices", header_style="cyan bold", show_lines=True
    )
    table.add_column("Addresses")
    table.add_column("MAC Address")

    def async_on_service_state_change(
        zeroconf: Zeroconf,
        service_type: str,
        name: str,
        state_change: ServiceStateChange,
    ) -> None:
        """Handle service state changes."""
        if state_change is not ServiceStateChange.Added:
            return

        future = asyncio.ensure_future(
            async_display_service_info(zeroconf, service_type, name)
        )
        background_tasks.add(future)
        future.add_done_callback(background_tasks.discard)

    async def async_display_service_info(
        zeroconf: Zeroconf, service_type: str, name: str
    ) -> None:
        """Retrieve and display service info."""
        info = AsyncServiceInfo(service_type, name)
        await info.async_request(zeroconf, 3000)
        if info is None:
            return

        console.print(f"[cyan bold]Found service {info.server}: is a WLED device 🎉")

        table.add_row(
            f"{str(info.server).rstrip('.')}\n"
            + ", ".join(info.parsed_scoped_addresses()),
            info.properties[b"mac"].decode(),  # type: ignore[union-attr]
        )

    console.print("[green]Scanning for WLED devices...")
    console.print("[red]Press Ctrl-C to exit\n")

    with Live(table, console=console, refresh_per_second=4):
        browser = AsyncServiceBrowser(
            zeroconf.zeroconf,
            "_wled._tcp.local.",
            handlers=[async_on_service_state_change],
        )

        try:
            while True:
                await asyncio.sleep(0.5)
        except KeyboardInterrupt:
            pass
        finally:
            console.print("\n[green]Control-C pressed, stopping scan")
            await browser.async_cancel()
            await zeroconf.async_close()


if __name__ == "__main__":
    cli()



"""Asynchronous Python client for WLED."""

from enum import IntEnum, IntFlag

from awesomeversion import AwesomeVersion

MIN_REQUIRED_VERSION = AwesomeVersion("0.14.0")


class LightCapability(IntFlag):
    """Enumeration representing the capabilities of a light in WLED."""

    NONE = 0
    RGB_COLOR = 1
    WHITE_CHANNEL = 2
    COLOR_TEMPERATURE = 4
    MANUAL_WHITE = 8

    # These are not used, but are reserved for future use.
    # WLED specifications documents we should expect them,
    # therefore, we include them here.
    RESERVED_2 = 16
    RESERVED_3 = 32
    RESERVED_4 = 64
    RESERVED_5 = 128


class LiveDataOverride(IntEnum):
    """Enumeration representing live override mode from WLED."""

    OFF = 0
    ON = 1
    OFF_UNTIL_REBOOT = 2


class NightlightMode(IntEnum):
    """Enumeration representing nightlight mode from WLED."""

    INSTANT = 0
    FADE = 1
    COLOR_FADE = 2
    SUNRISE = 3


class SyncGroup(IntFlag):
    """Bitfield for udp sync groups 1-8."""

    NONE = 0
    GROUP1 = 1
    GROUP2 = 2
    GROUP3 = 4
    GROUP4 = 8
    GROUP5 = 16
    GROUP6 = 32
    GROUP7 = 64
    GROUP8 = 128



"""Asynchronous Python client for WLED.

Adaptation of the snippet/code from:
- https://github.com/tiangolo/typer/issues/88#issuecomment-1613013597
- https://github.com/argilla-io/argilla/blob/e77ca86c629a492019f230ac55ebde207b280xc9c/src/argilla/cli/typer_ext.py
"""

#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from __future__ import annotations

import asyncio
from functools import wraps
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Coroutine,
    ParamSpec,
    TypeVar,
)

from typer import Exit
from typer import Typer as SyncTyper

if TYPE_CHECKING:
    from typer.core import TyperCommand, TyperGroup

_P = ParamSpec("_P")
_R = TypeVar("_R")

HandleErrorFunc = Callable[[Any], None]


class AsyncTyper(SyncTyper):
    """A Typer subclass that supports async."""

    error_handlers: dict[type[Exception], HandleErrorFunc]

    # pylint: disable-next=too-many-arguments, too-many-locals
    def callback(  # type: ignore[override] # noqa: PLR0913
        self,
        name: str | None = None,
        *,
        cls: type[TyperGroup] | None = None,
        invoke_without_command: bool = False,
        no_args_is_help: bool = False,
        subcommand_metavar: str | None = None,
        chain: bool = False,
        result_callback: Callable[..., Any] | None = None,
        context_settings: dict[Any, Any] | None = None,
        # pylint: disable-next=redefined-builtin
        help: str | None = None,  # noqa: A002
        epilog: str | None = None,
        short_help: str | None = None,
        options_metavar: str = "[OPTIONS]",
        add_help_option: bool = True,
        hidden: bool = False,
        deprecated: bool = False,
        rich_help_panel: str | None = None,
    ) -> Callable[
        [Callable[_P, Coroutine[Any, Any, _R]]],
        Callable[_P, Coroutine[Any, Any, _R]],
    ]:
        """Create a new typer callback."""
        super_callback = super().callback(
            name,
            cls=cls,
            invoke_without_command=invoke_without_command,
            no_args_is_help=no_args_is_help,
            subcommand_metavar=subcommand_metavar,
            chain=chain,
            result_callback=result_callback,
            context_settings=context_settings,
            help=help,
            epilog=epilog,
            short_help=short_help,
            options_metavar=options_metavar,
            add_help_option=add_help_option,
            hidden=hidden,
            deprecated=deprecated,
            rich_help_panel=rich_help_panel,
        )

        def decorator(
            func: Callable[_P, Coroutine[Any, Any, _R]],
        ) -> Callable[_P, Coroutine[Any, Any, _R]]:
            if asyncio.iscoroutinefunction(func):

                @wraps(func)
                def sync_func(*_args: _P.args, **_kwargs: _P.kwargs) -> _R:
                    return asyncio.run(func(*_args, **_kwargs))

                super_callback(sync_func)
            else:
                super_callback(func)

            return func

        return decorator

    # pylint: disable-next=too-many-arguments
    def command(  # type: ignore[override] # noqa: PLR0913
        self,
        name: str | None = None,
        *,
        cls: type[TyperCommand] | None = None,
        context_settings: dict[Any, Any] | None = None,
        # pylint: disable-next=redefined-builtin
        help: str | None = None,  # noqa: A002
        epilog: str | None = None,
        short_help: str | None = None,
        options_metavar: str = "[OPTIONS]",
        add_help_option: bool = True,
        no_args_is_help: bool = False,
        hidden: bool = False,
        deprecated: bool = False,
        # Rich settings
        rich_help_panel: str | None = None,
    ) -> Callable[
        [Callable[_P, Coroutine[Any, Any, _R]]],
        Callable[_P, Coroutine[Any, Any, _R]],
    ]:
        """Create a new typer command."""
        super_command = super().command(
            name,
            cls=cls,
            context_settings=context_settings,
            help=help,
            epilog=epilog,
            short_help=short_help,
            options_metavar=options_metavar,
            add_help_option=add_help_option,
            no_args_is_help=no_args_is_help,
            hidden=hidden,
            deprecated=deprecated,
            rich_help_panel=rich_help_panel,
        )

        def decorator(
            func: Callable[_P, Coroutine[Any, Any, _R]],
        ) -> Callable[_P, Coroutine[Any, Any, _R]]:
            if asyncio.iscoroutinefunction(func):

                @wraps(func)
                def sync_func(*_args: _P.args, **_kwargs: _P.kwargs) -> _R:
                    return asyncio.run(func(*_args, **_kwargs))

                super_command(sync_func)
            else:
                super_command(func)

            return func

        return decorator

    def error_handler(self, exc: type[Exception]) -> Callable[[HandleErrorFunc], None]:
        """Register an error handler for a given exception."""
        if not hasattr(self, "error_handlers"):
            self.error_handlers = {}

        def decorator(func: HandleErrorFunc) -> None:
            self.error_handlers[exc] = func

        return decorator

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        """Call the typer app."""
        try:
            return super().__call__(*args, **kwargs)
        except Exit:
            raise
        # pylint: disable-next=broad-except
        except Exception as e:
            if (
                not hasattr(self, "error_handlers")
                or (handler := self.error_handlers.get(type(e))) is None
            ):
                raise
            return handler(e)



"""Exceptions for WLED."""


class WLEDError(Exception):
    """Generic WLED exception."""


class WLEDEmptyResponseError(Exception):
    """WLED empty API response exception."""


class WLEDConnectionError(WLEDError):
    """WLED connection exception."""


class WLEDConnectionTimeoutError(WLEDConnectionError):
    """WLED connection Timeout exception."""


class WLEDConnectionClosedError(WLEDConnectionError):
    """WLED WebSocket connection has been closed."""


class WLEDUnsupportedVersionError(WLEDError):
    """WLED version is unsupported."""


class WLEDUpgradeError(WLEDError):
    """WLED upgrade exception."""



"""Asynchronous Python client for WLED."""

from .const import LightCapability, LiveDataOverride, NightlightMode, SyncGroup
from .exceptions import (
    WLEDConnectionClosedError,
    WLEDConnectionError,
    WLEDConnectionTimeoutError,
    WLEDError,
    WLEDUnsupportedVersionError,
    WLEDUpgradeError,
)
from .models import (
    Device,
    Effect,
    Info,
    Leds,
    Nightlight,
    Palette,
    Playlist,
    PlaylistEntry,
    Preset,
    Releases,
    Segment,
    State,
    UDPSync,
)
from .wled import WLED, WLEDReleases

__all__ = [
    "Device",
    "Effect",
    "Info",
    "Leds",
    "LightCapability",
    "LiveDataOverride",
    "Nightlight",
    "NightlightMode",
    "Palette",
    "Playlist",
    "PlaylistEntry",
    "Preset",
    "Releases",
    "Segment",
    "State",
    "SyncGroup",
    "UDPSync",
    "WLED",
    "WLEDConnectionClosedError",
    "WLEDConnectionError",
    "WLEDConnectionTimeoutError",
    "WLEDError",
    "WLEDReleases",
    "WLEDUnsupportedVersionError",
    "WLEDUpgradeError",
]



"""Asynchronous Python client for WLED."""

from __future__ import annotations

import asyncio
import socket
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Self

import aiohttp
import backoff
import orjson
from yarl import URL

from .exceptions import (
    WLEDConnectionClosedError,
    WLEDConnectionError,
    WLEDConnectionTimeoutError,
    WLEDEmptyResponseError,
    WLEDError,
    WLEDUpgradeError,
)
from .models import Device, Playlist, Preset, Releases

if TYPE_CHECKING:
    from collections.abc import Callable, Sequence

    from awesomeversion import AwesomeVersion

    from .const import LiveDataOverride


@dataclass
class WLED:
    """Main class for handling connections with WLED."""

    host: str
    request_timeout: float = 8.0
    session: aiohttp.client.ClientSession | None = None

    _client: aiohttp.ClientWebSocketResponse | None = None
    _close_session: bool = False
    _device: Device | None = None

    @property
    def connected(self) -> bool:
        """Return if we are connect to the WebSocket of a WLED device.

        Returns
        -------
            True if we are connected to the WebSocket of a WLED device,
            False otherwise.

        """
        return self._client is not None and not self._client.closed

    async def connect(self) -> None:
        """Connect to the WebSocket of a WLED device.

        Raises
        ------
            WLEDError: The configured WLED device, does not support WebSocket
                communications.
            WLEDConnectionError: Error occurred while communicating with
                the WLED device via the WebSocket.

        """
        if self.connected:
            return

        if not self._device:
            await self.update()

        if not self.session or not self._device or self._device.info.websocket is None:
            msg = "The WLED device at {self.host} does not support WebSockets"
            raise WLEDError(msg)

        url = URL.build(scheme="ws", host=self.host, port=80, path="/ws")

        try:
            self._client = await self.session.ws_connect(url=url, heartbeat=30)
        except (
            aiohttp.WSServerHandshakeError,
            aiohttp.ClientConnectionError,
            socket.gaierror,
        ) as exception:
            msg = (
                "Error occurred while communicating with WLED device"
                f" on WebSocket at {self.host}"
            )
            raise WLEDConnectionError(msg) from exception

    async def listen(self, callback: Callable[[Device], None]) -> None:
        """Listen for events on the WLED WebSocket.

        Args:
        ----
            callback: Method to call when a state update is received from
                the WLED device.

        Raises:
        ------
            WLEDError: Not connected to a WebSocket.
            WLEDConnectionError: An connection error occurred while connected
                to the WLED device.
            WLEDConnectionClosedError: The WebSocket connection to the remote WLED
                has been closed.

        """
        if not self._client or not self.connected or not self._device:
            msg = "Not connected to a WLED WebSocket"
            raise WLEDError(msg)

        while not self._client.closed:
            message = await self._client.receive()

            if message.type == aiohttp.WSMsgType.ERROR:
                raise WLEDConnectionError(self._client.exception())

            if message.type == aiohttp.WSMsgType.TEXT:
                message_data = message.json()
                device = self._device.update_from_dict(data=message_data)
                callback(device)

            if message.type in (
                aiohttp.WSMsgType.CLOSE,
                aiohttp.WSMsgType.CLOSED,
                aiohttp.WSMsgType.CLOSING,
            ):
                msg = f"Connection to the WLED WebSocket on {self.host} has been closed"
                raise WLEDConnectionClosedError(msg)

    async def disconnect(self) -> None:
        """Disconnect from the WebSocket of a WLED device."""
        if not self._client or not self.connected:
            return

        await self._client.close()

    @backoff.on_exception(backoff.expo, WLEDConnectionError, max_tries=3, logger=None)
    async def request(
        self,
        uri: str = "",
        method: str = "GET",
        data: dict[str, Any] | None = None,
    ) -> Any:
        """Handle a request to a WLED device.

        A generic method for sending/handling HTTP requests done gainst
        the WLED device.

        Args:
        ----
            uri: Request URI, for example `/json/si`.
            method: HTTP method to use for the request.E.g., "GET" or "POST".
            data: Dictionary of data to send to the WLED device.

        Returns:
        -------
            A Python dictionary (JSON decoded) with the response from the
            WLED device.

        Raises:
        ------
            WLEDConnectionError: An error occurred while communication with
                the WLED device.
            WLEDConnectionTimeoutError: A timeout occurred while communicating
                with the WLED device.
            WLEDError: Received an unexpected response from the WLED device.

        """
        url = URL.build(scheme="http", host=self.host, port=80, path=uri)

        headers = {
            "Accept": "application/json, text/plain, */*",
        }

        if self.session is None:
            self.session = aiohttp.ClientSession()
            self._close_session = True

        # If updating the state, always request for a state response
        if method == "POST" and uri == "/json/state" and data is not None:
            data["v"] = True

        try:
            async with asyncio.timeout(self.request_timeout):
                response = await self.session.request(
                    method,
                    url,
                    json=data,
                    headers=headers,
                )

            content_type = response.headers.get("Content-Type", "")
            if response.status // 100 in [4, 5]:
                contents = await response.read()
                response.close()

                if content_type == "application/json":
                    raise WLEDError(
                        response.status,
                        orjson.loads(contents),
                    )
                raise WLEDError(
                    response.status,
                    {"message": contents.decode("utf8")},
                )

            response_data = await response.text()
            if "application/json" in content_type:
                response_data = orjson.loads(response_data)

        except asyncio.TimeoutError as exception:
            msg = f"Timeout occurred while connecting to WLED device at {self.host}"
            raise WLEDConnectionTimeoutError(msg) from exception
        except (aiohttp.ClientError, socket.gaierror) as exception:
            msg = f"Error occurred while communicating with WLED device at {self.host}"
            raise WLEDConnectionError(msg) from exception

        if "application/json" in content_type and (
            method == "POST"
            and uri == "/json/state"
            and self._device is not None
            and data is not None
        ):
            self._device.update_from_dict(data={"state": response_data})

        return response_data

    @backoff.on_exception(
        backoff.expo,
        WLEDEmptyResponseError,
        max_tries=3,
        logger=None,
    )
    async def update(self) -> Device:
        """Get all information about the device in a single call.

        This method updates all WLED information available with a single API
        call.

        Returns
        -------
            WLED Device data.

        Raises
        ------
            WLEDEmptyResponseError: The WLED device returned an empty response.

        """
        if not (data := await self.request("/json")):
            msg = (
                f"WLED device at {self.host} returned an empty API"
                " response on full update",
            )
            raise WLEDEmptyResponseError(msg)

        if not (presets := await self.request("/presets.json")):
            msg = (
                f"WLED device at {self.host} returned an empty API"
                " response on presets update",
            )
            raise WLEDEmptyResponseError(msg)
        data["presets"] = presets

        if not self._device:
            self._device = Device.from_dict(data)
        else:
            self._device.update_from_dict(data)

        return self._device

    async def master(
        self,
        *,
        brightness: int | None = None,
        on: bool | None = None,
        transition: int | None = None,
    ) -> None:
        """Change master state of a WLED Light device.

        Args:
        ----
            brightness: The brightness of the light master, between 0 and 255.
            on: A boolean, true to turn the master light on, false otherwise.
            transition: Duration of the crossfade between different
                colors/brightness levels. One unit is 100ms, so a value of 4
                results in a transition of 400ms.

        """
        state: dict[str, bool | int] = {}

        if brightness is not None:
            state["bri"] = brightness

        if on is not None:
            state["on"] = on

        if transition is not None:
            state["tt"] = transition

        await self.request("/json/state", method="POST", data=state)

    # pylint: disable=too-many-locals, too-many-branches, too-many-arguments
    async def segment(  # noqa: PLR0912, PLR0913
        self,
        segment_id: int,
        *,
        brightness: int | None = None,
        clones: int | None = None,
        color_primary: tuple[int, int, int, int] | tuple[int, int, int] | None = None,
        color_secondary: tuple[int, int, int, int] | tuple[int, int, int] | None = None,
        color_tertiary: tuple[int, int, int, int] | tuple[int, int, int] | None = None,
        effect: int | str | None = None,
        individual: Sequence[
            int | Sequence[int] | tuple[int, int, int, int] | tuple[int, int, int]
        ]
        | None = None,
        intensity: int | None = None,
        length: int | None = None,
        on: bool | None = None,
        palette: int | str | None = None,
        reverse: bool | None = None,
        selected: bool | None = None,
        speed: int | None = None,
        start: int | None = None,
        stop: int | None = None,
        transition: int | None = None,
        cct: int | None = None,
    ) -> None:
        """Change state of a WLED Light segment.

        Args:
        ----
            segment_id: The ID of the segment to adjust.
            brightness: The brightness of the segment, between 0 and 255.
            clones: Deprecated.
            color_primary: The primary color of this segment.
            color_secondary: The secondary color of this segment.
            color_tertiary: The tertiary color of this segment.
            effect: The effect number (or name) to use on this segment.
            individual: A list of colors to use for each LED in the segment.
            intensity: The effect intensity to use on this segment.
            length: The length of this segment.
            on: A boolean, true to turn this segment on, false otherwise.
            palette: the palette number or name to use on this segment.
            reverse: Flips the segment, causing animations to change direction.
            selected: Selected segments will have their state (color/FX) updated
                by APIs that don't support segments.
            speed: The relative effect speed, between 0 and 255.
            start: LED the segment starts at.
            stop: LED the segment stops at, not included in range. If stop is
                set to a lower or equal value than start (setting to 0 is
                recommended), the segment is invalidated and deleted.
            transition:  Duration of the crossfade between different
                colors/brightness levels. One unit is 100ms, so a value of 4
                results in a transition of 400ms.
            cct: White spectrum color temperature.

        Raises:
        ------
            WLEDError: Something went wrong setting the segment state.

        """
        if self._device is None:
            await self.update()

        if self._device is None:
            msg = "Unable to communicate with WLED to get the current state"
            raise WLEDError(msg)

        state = {}  # type: ignore[var-annotated]
        segment = {
            "bri": brightness,
            "cln": clones,
            "fx": effect,
            "i": individual,
            "ix": intensity,
            "len": length,
            "on": on,
            "pal": palette,
            "rev": reverse,
            "sel": selected,
            "start": start,
            "stop": stop,
            "sx": speed,
            "cct": cct,
        }

        # Find effect if it was based on a name
        if effect is not None and isinstance(effect, str):
            segment["fx"] = next(
                (
                    item.effect_id
                    for item in self._device.effects.values()
                    if item.name.lower() == effect.lower()
                ),
                None,
            )

        # Find palette if it was based on a name
        if palette is not None and isinstance(palette, str):
            segment["pal"] = next(
                (
                    item.palette_id
                    for item in self._device.palettes.values()
                    if item.name.lower() == palette.lower()
                ),
                None,
            )

        # Filter out not set values
        state = {k: v for k, v in state.items() if v is not None}
        segment = {k: v for k, v in segment.items() if v is not None}

        # Determine color set
        colors = []
        if color_primary is not None:
            colors.append(color_primary)
        elif color_secondary is not None or color_tertiary is not None:
            if clrs := self._device.state.segments[segment_id].color:
                colors.append(clrs.primary)
            else:
                colors.append((0, 0, 0))

        if color_secondary is not None:
            colors.append(color_secondary)
        elif color_tertiary is not None:
            if (
                clrs := self._device.state.segments[segment_id].color
            ) and clrs.secondary:
                colors.append(clrs.secondary)
            else:
                colors.append((0, 0, 0))

        if color_tertiary is not None:
            colors.append(color_tertiary)

        if colors:
            segment["col"] = colors

        if segment:
            segment["id"] = segment_id
            state["seg"] = [segment]

        if transition is not None:
            state["tt"] = transition

        await self.request("/json/state", method="POST", data=state)

    async def transition(self, transition: int) -> None:
        """Set the default transition time for manual control.

        Args:
        ----
            transition: Duration of the default crossfade between different
                colors/brightness levels. One unit is 100ms, so a value of 4
                results in a transition of 400ms.

        """
        await self.request(
            "/json/state",
            method="POST",
            data={"transition": transition},
        )

    async def preset(self, preset: int | str | Preset) -> None:
        """Set a preset on a WLED device.

        Args:
        ----
            preset: The preset to activate on this WLED device.

        """
        # Find preset if it was based on a name
        if self._device and self._device.presets and isinstance(preset, str):
            preset = next(
                (
                    item.preset_id
                    for item in self._device.presets.values()
                    if item.name.lower() == preset.lower()
                ),
                preset,
            )

        if isinstance(preset, Preset):
            preset = preset.preset_id

        await self.request("/json/state", method="POST", data={"ps": preset})

    async def playlist(self, playlist: int | str | Playlist) -> None:
        """Set a playlist on a WLED device.

        Args:
        ----
            playlist: The playlist to activate on this WLED device.

        """
        # Find playlist if it was based on a name
        if self._device and self._device.playlists and isinstance(playlist, str):
            playlist = next(
                (
                    item.playlist_id
                    for item in self._device.playlists.values()
                    if item.name.lower() == playlist.lower()
                ),
                playlist,
            )

        if isinstance(playlist, Playlist):
            playlist = playlist.playlist_id

        await self.request("/json/state", method="POST", data={"ps": playlist})

    async def live(self, live: LiveDataOverride) -> None:
        """Set the live override mode on a WLED device.

        Args:
        ----
            live: The live override mode to set on this WLED device.

        """
        await self.request("/json/state", method="POST", data={"lor": live.value})

    async def sync(
        self,
        *,
        send: bool | None = None,
        receive: bool | None = None,
    ) -> None:
        """Set the sync status of the WLED device.

        Args:
        ----
            send: Send WLED broadcast (UDP sync) packet on state change.
            receive: Receive broadcast packets.

        """
        sync = {"send": send, "recv": receive}
        sync = {k: v for k, v in sync.items() if v is not None}
        await self.request("/json/state", method="POST", data={"udpn": sync})

    async def nightlight(
        self,
        *,
        duration: int | None = None,
        fade: bool | None = None,
        on: bool | None = None,
        target_brightness: int | None = None,
    ) -> None:
        """Control the nightlight function of a WLED device.

        Args:
        ----
            duration: Duration of nightlight in minutes.
            fade: If true, the light will gradually dim over the course of the
                nightlight duration. If false, it will instantly turn to the
                target brightness once the duration has elapsed.
            on: A boolean, true to turn the nightlight on, false otherwise.
            target_brightness: Target brightness of nightlight, between 0 and 255.

        """
        nightlight = {
            "dur": duration,
            "fade": fade,
            "on": on,
            "tbri": target_brightness,
        }

        # Filter out not set values
        nightlight = {k: v for k, v in nightlight.items() if v is not None}
        await self.request("/json/state", method="POST", data={"nl": nightlight})

    async def upgrade(self, *, version: str | AwesomeVersion) -> None:
        """Upgrades WLED device to the specified version.

        Args:
        ----
            version: The version to upgrade to.

        Raises:
        ------
            WLEDUpgradeError: If the upgrade has failed.
            WLEDConnectionTimeoutError: When a connection timeout occurs.
            WLEDConnectionError: When a connection error occurs.

        """
        if self._device is None:
            await self.update()

        if self.session is None or self._device is None:
            msg = "Unexpected upgrade error; No session or device"
            raise WLEDUpgradeError(msg)

        if self._device.info.architecture not in {
            "esp01",
            "esp02",
            "esp32",
            "esp8266",
            "esp32-c3",
            "esp32-s2",
            "esp32-s3",
        }:
            msg = (
                "Upgrade is only supported on ESP01, ESP02, ESP32, ESP8266, "
                "ESP32-C3, ESP32-S2, and ESP32-S3 devices"
            )
            raise WLEDUpgradeError(msg)

        if not self._device.info.version:
            msg = "Current version is unknown, cannot perform upgrade"
            raise WLEDUpgradeError(msg)

        if self._device.info.version == version:
            msg = "Device already running the requested version"
            raise WLEDUpgradeError(msg)

        # Determine if this is an Ethernet board
        ethernet = ""
        if (
            self._device.info.architecture == "esp32"
            and self._device.info.wifi is not None
            and not self._device.info.wifi.bssid
            and self._device.info.version
            and self._device.info.version >= "0.10.0"
        ):
            ethernet = "_Ethernet"

        # Determine if this is a 2M ESP8266 board.
        # See issue `https://github.com/Aircoookie/WLED/issues/3257`
        gzip = ""
        if self._device.info.architecture == "esp02":
            gzip = ".gz"

        url = URL.build(scheme="http", host=self.host, port=80, path="/update")
        architecture = self._device.info.architecture.upper()
        update_file = f"WLED_{version}_{architecture}{ethernet}.bin{gzip}"
        download_url = (
            "https://github.com/Aircoookie/WLED/releases/download"
            f"/v{version}/{update_file}"
        )

        try:
            async with asyncio.timeout(
                self.request_timeout * 10,
            ), self.session.get(
                download_url,
                raise_for_status=True,
            ) as download:
                form = aiohttp.FormData()
                form.add_field("file", await download.read(), filename=update_file)
                await self.session.post(url, data=form)
        except asyncio.TimeoutError as exception:
            msg = "Timeout occurred while fetching WLED version information from GitHub"
            raise WLEDConnectionTimeoutError(msg) from exception
        except aiohttp.ClientResponseError as exception:
            if exception.status == 404:
                msg = f"Requested WLED version '{version}' does not exists"
                raise WLEDUpgradeError(msg) from exception
            msg = (
                f"Could not download requested WLED version '{version}'"
                f" from {download_url}"
            )
            raise WLEDUpgradeError(msg) from exception
        except (aiohttp.ClientError, socket.gaierror) as exception:
            msg = (
                "Timeout occurred while communicating with GitHub"
                " for WLED version information"
            )
            raise WLEDConnectionError(msg) from exception

    async def reset(self) -> None:
        """Reboot WLED device."""
        await self.request("/reset")

    async def close(self) -> None:
        """Close open client (WebSocket) session."""
        await self.disconnect()
        if self.session and self._close_session:
            await self.session.close()

    async def __aenter__(self) -> Self:
        """Async enter.

        Returns
        -------
            The WLED object.

        """
        return self

    async def __aexit__(self, *_exc_info: object) -> None:
        """Async exit.

        Args:
        ----
            _exc_info: Exec type.

        """
        await self.close()


@dataclass
class WLEDReleases:
    """Get version information for WLED."""

    request_timeout: float = 8.0
    session: aiohttp.client.ClientSession | None = None

    _client: aiohttp.ClientWebSocketResponse | None = None
    _close_session: bool = False

    @backoff.on_exception(backoff.expo, WLEDConnectionError, max_tries=3, logger=None)
    async def releases(self) -> Releases:
        """Fetch WLED version information from GitHub.

        Returns
        -------
            A dictionary of WLED versions, with the key being the version type.

        Raises
        ------
            WLEDConnectionTimeoutError: Timeout occurred while fetching WLED
                version information from GitHub.
            WLEDConnectionError: Timeout occurred while communicating with
                GitHub for WLED version information.
            WLEDError: Didn't get a JSON response from GitHub while retrieving
                version information.

        """
        if self.session is None:
            self.session = aiohttp.ClientSession()
            self._close_session = True

        try:
            async with asyncio.timeout(self.request_timeout):
                response = await self.session.get(
                    "https://api.github.com/repos/Aircoookie/WLED/releases",
                    headers={"Accept": "application/json"},
                )
        except asyncio.TimeoutError as exception:
            msg = (
                "Timeout occurred while fetching WLED releases information from GitHub"
            )
            raise WLEDConnectionTimeoutError(msg) from exception
        except (aiohttp.ClientError, socket.gaierror) as exception:
            msg = "Timeout occurred while communicating with GitHub for WLED releases"
            raise WLEDConnectionError(msg) from exception

        content_type = response.headers.get("Content-Type", "")
        contents = await response.read()
        if response.status // 100 in [4, 5]:
            response.close()

            if content_type == "application/json":
                raise WLEDError(response.status, orjson.loads(contents))
            raise WLEDError(response.status, {"message": contents.decode("utf8")})

        if "application/json" not in content_type:
            msg = "No JSON response from GitHub while retrieving WLED releases"
            raise WLEDError(msg)

        releases = orjson.loads(contents)
        version_latest = None
        version_latest_beta = None
        for release in releases:
            if (
                release["prerelease"] is False
                and "b" not in release["tag_name"].lower()
                and version_latest is None
            ):
                version_latest = release["tag_name"].lstrip("vV")
            if (
                release["prerelease"] is True or "b" in release["tag_name"].lower()
            ) and version_latest_beta is None:
                version_latest_beta = release["tag_name"].lstrip("vV")
            if version_latest is not None and version_latest_beta is not None:
                break

        return Releases(
            beta=version_latest_beta,
            stable=version_latest,
        )

    async def close(self) -> None:
        """Close open client session."""
        if self.session and self._close_session:
            await self.session.close()

    async def __aenter__(self) -> Self:
        """Async enter.

        Returns
        -------
            The WLEDReleases object.

        """
        return self

    async def __aexit__(self, *_exc_info: object) -> None:
        """Async exit.

        Args:
        ----
            _exc_info: Exec type.

        """
        await self.close()



"""Models for WLED."""

from __future__ import annotations

from dataclasses import dataclass, field
from datetime import UTC, datetime, timedelta
from functools import cached_property
from typing import Any

from awesomeversion import AwesomeVersion
from mashumaro import field_options
from mashumaro.config import BaseConfig
from mashumaro.mixins.orjson import DataClassORJSONMixin
from mashumaro.types import SerializableType, SerializationStrategy

from .const import (
    MIN_REQUIRED_VERSION,
    LightCapability,
    LiveDataOverride,
    NightlightMode,
    SyncGroup,
)
from .exceptions import WLEDUnsupportedVersionError
from .utils import get_awesome_version


class AwesomeVersionSerializationStrategy(SerializationStrategy, use_annotations=True):
    """Serialization strategy for AwesomeVersion objects."""

    def serialize(self, value: AwesomeVersion | None) -> str:
        """Serialize AwesomeVersion object to string."""
        if value is None:
            return ""
        return str(value)

    def deserialize(self, value: str) -> AwesomeVersion | None:
        """Deserialize string to AwesomeVersion object."""
        version = get_awesome_version(value)
        if not version.valid:
            return None
        return version


class TimedeltaSerializationStrategy(SerializationStrategy, use_annotations=True):
    """Serialization strategy for timedelta objects."""

    def serialize(self, value: timedelta) -> int:
        """Serialize timedelta object to seconds."""
        return int(value.total_seconds())

    def deserialize(self, value: int) -> timedelta:
        """Deserialize integer to timedelta object."""
        return timedelta(seconds=value)


class TimestampSerializationStrategy(SerializationStrategy, use_annotations=True):
    """Serialization strategy for datetime objects."""

    def serialize(self, value: datetime) -> float:
        """Serialize datetime object to timestamp."""
        return value.timestamp()

    def deserialize(self, value: float) -> datetime:
        """Deserialize timestamp to datetime object."""
        return datetime.fromtimestamp(value, tz=UTC)


@dataclass
class Color(SerializableType):
    """Object holding color information in WLED."""

    primary: tuple[int, int, int, int] | tuple[int, int, int]
    secondary: tuple[int, int, int, int] | tuple[int, int, int] | None = None
    tertiary: tuple[int, int, int, int] | tuple[int, int, int] | None = None

    def _serialize(self) -> list[tuple[int, int, int, int] | tuple[int, int, int]]:
        colors = [self.primary]
        if self.secondary is not None:
            colors.append(self.secondary)
            if self.tertiary is not None:
                colors.append(self.tertiary)
        return colors

    @classmethod
    def _deserialize(
        cls, value: list[tuple[int, int, int, int] | tuple[int, int, int] | str]
    ) -> Color:
        # Some values in the list can be strings, which indicates that the
        # color is a hex color value.
        return cls(
            *[  # type: ignore[arg-type]
                tuple(int(color[i : i + 2], 16) for i in (1, 3, 5))
                if isinstance(color, str)
                else color
                for color in value
            ]
        )


class BaseModel(DataClassORJSONMixin):
    """Base model for all WLED models."""

    # pylint: disable-next=too-few-public-methods
    class Config(BaseConfig):
        """Mashumaro configuration."""

        omit_none = True
        serialization_strategy = {  # noqa: RUF012
            AwesomeVersion: AwesomeVersionSerializationStrategy(),
            datetime: TimestampSerializationStrategy(),
            timedelta: TimedeltaSerializationStrategy(),
        }
        serialize_by_alias = True


@dataclass(kw_only=True)
class Nightlight(BaseModel):
    """Object holding nightlight state in WLED."""

    duration: int = field(default=1, metadata=field_options(alias="dur"))
    """Duration of nightlight in minutes."""

    mode: NightlightMode = field(default=NightlightMode.INSTANT)
    """Nightlight mode (available since 0.10.2)."""

    on: bool = field(default=False)
    """Nightlight currently active."""

    target_brightness: int = field(default=0, metadata=field_options(alias="tbri"))
    """Target brightness of nightlight feature."""


@dataclass(kw_only=True)
class UDPSync(BaseModel):
    """Object holding UDP sync state in WLED.

    Missing at this point, is the `nn` field. This field allows to skip
    sending a broadcast packet for the current API request; However, this field
    is only used for requests and not part of the state responses.
    """

    receive: bool = field(default=False, metadata=field_options(alias="recv"))
    """Receive broadcast packets."""

    receive_groups: SyncGroup = field(
        default=SyncGroup.NONE, metadata=field_options(alias="rgrp")
    )
    """Groups to receive WLED broadcast packets from."""

    send: bool = field(default=False, metadata=field_options(alias="send"))
    """Send WLED broadcast (UDP sync) packet on state change."""

    send_groups: SyncGroup = field(
        default=SyncGroup.NONE, metadata=field_options(alias="sgrp")
    )
    """Groups to send WLED broadcast packets to."""


@dataclass(frozen=True, kw_only=True)
class Effect(BaseModel):
    """Object holding an effect in WLED."""

    effect_id: int
    name: str


@dataclass(frozen=True, kw_only=True)
class Palette(BaseModel):
    """Object holding an palette in WLED.

    Args:
    ----
        data: The data from the WLED device API.

    Returns:
    -------
        A palette object.

    """

    name: str
    palette_id: int


@dataclass(kw_only=True)
class Segment(BaseModel):
    """Object holding segment state in WLED.

    Args:
    ----
        data: The data from the WLED device API.

    Returns:
    -------
        A segment object.

    """

    brightness: int = field(default=0, metadata=field_options(alias="bri"))
    """Brightness of the segment."""

    clones: int = field(default=-1, metadata=field_options(alias="cln"))
    """The segment this segment clones."""

    color: Color | None = field(default=None, metadata=field_options(alias="col"))
    """The primary, secondary (background) and tertiary colors of the segment.

    Each color is an tuple of 3 or 4 bytes, which represents a RGB(W) color,
    i.e. (255,170,0) or (64,64,64,64).

    WLED can also return hex color values as strings, this library will
    automatically convert those to RGB values to keep the data consistent.
    """

    effect_id: int | str = field(default=0, metadata=field_options(alias="fx"))
    """ID of the effect.

    ~ to increment, ~- to decrement, or "r" for random.
    """

    intensity: int | str = field(default=0, metadata=field_options(alias="ix"))
    """Intensity of the segment.

    Effect intensity. ~ to increment, ~- to decrement. ~10 to increment by 10,
    ~-10 to decrement by 10.
    """

    length: int = field(default=0, metadata=field_options(alias="len"))
    """Length of the segment (stop - start).

    Stop has preference, so if it is included, length is ignored.
    """

    on: bool | None = field(default=None)
    """The on/off state of the segment."""

    palette_id: int | str = field(default=0, metadata=field_options(alias="pal"))
    """ID of the palette.

    ~ to increment, ~- to decrement, or r for random.
    """

    reverse: bool = field(default=False, metadata=field_options(alias="rev"))
    """
    Flips the segment (in horizontal dimension for 2D set-up),
    causing animations to change direction.
    """

    segment_id: int | None = field(default=None, metadata=field_options(alias="id"))
    """The ID of the segment."""

    selected: bool = field(default=False, metadata=field_options(alias="sel"))
    """
    Indicates if the segment is selected.

    Selected segments will have their state (color/FX) updated by APIs that
    don't support segments (e.g. UDP sync, HTTP API). If no segment is selected,
    the first segment (id:0) will behave as if selected.

    WLED will report the state of the first (lowest id) segment that is selected
    to APIs (HTTP, MQTT, Blynk...), or mainseg in case no segment is selected
    and for the UDP API.

    Live data is always applied to all LEDs regardless of segment configuration.
    """

    speed: int = field(default=0, metadata=field_options(alias="sx"))
    """Relative effect speed.

    ~ to increment, ~- to decrement. ~10 to increment by 10, ~-10 to decrement by 10.
    """

    start: int = 0
    """LED the segment starts at.

    For 2D set-up it determines column where segment starts,
    from top-left corner of the matrix.
    """

    stop: int = 0
    """LED the segment stops at, not included in range.

    If stop is set to a lower or equal value than start (setting to 0 is
    recommended), the segment is invalidated and deleted.

    For 2D set-up it determines column where segment stops,
    from top-left corner of the matrix.
    """

    cct: int = field(default=0)
    """White spectrum color temperature.

    0 indicates the warmest possible color temperature,
    255 indicates the coldest temperature
    """


@dataclass(kw_only=True)
class Leds:
    """Object holding leds info from WLED."""

    count: int = 0
    """Total LED count."""

    fps: int = 0
    """Current frames per second."""

    light_capabilities: LightCapability = field(
        default=LightCapability.NONE, metadata=field_options(alias="lc")
    )
    """Capabilities of the light."""

    max_power: int = field(default=0, metadata=field_options(alias="maxpwr"))
    """Maximum power budget in milliamps for the ABL. 0 if ABL is disabled."""

    max_segments: int = field(default=0, metadata=field_options(alias="maxseg"))
    """Maximum number of segments supported by this version."""

    power: int = field(default=0, metadata=field_options(alias="pwr"))
    """
    Current LED power usage in milliamps as determined by the ABL.
    0 if ABL is disabled.
    """

    segment_light_capabilities: list[LightCapability] = field(
        default_factory=list, metadata=field_options(alias="seglc")
    )
    """Capabilities of each segment."""


@dataclass(kw_only=True)
class Wifi(BaseModel):
    """Object holding Wi-Fi information from WLED.

    Args:
    ----
        data: The data from the WLED device API.

    Returns:
    -------
        A Wi-Fi object.

    """

    bssid: str = "00:00:00:00:00:00"
    channel: int = 0
    rssi: int = 0
    signal: int = 0


@dataclass(kw_only=True)
class Filesystem(BaseModel):
    """Object holding Filesystem information from WLED.

    Args:
    ----
        data: The data from the WLED device API.

    Returns:
    -------
        A Filesystem object.

    """

    last_modified: datetime | None = field(
        default=None, metadata=field_options(alias="pmt")
    )
    """
    Last modification of the presets.json file. Not accurate after boot or
    after using /edit.
    """

    total: int = field(default=1, metadata=field_options(alias="t"))
    """Total space of the filesystem in kilobytes."""

    used: int = field(default=1, metadata=field_options(alias="u"))
    """Used space of the filesystem in kilobytes."""

    @cached_property
    def free(self) -> int:
        """Return the free space of the filesystem in kilobytes.

        Returns
        -------
            The free space of the filesystem.

        """
        return self.total - self.used

    @cached_property
    def free_percentage(self) -> int:
        """Return the free percentage of the filesystem.

        Returns
        -------
            The free percentage of the filesystem.

        """
        return round((self.free / self.total) * 100)

    @cached_property
    def used_percentage(self) -> int:
        """Return the used percentage of the filesystem.

        Returns
        -------
            The used percentage of the filesystem.

        """
        return round((self.used / self.total) * 100)


@dataclass(kw_only=True)
class Info(BaseModel):  # pylint: disable=too-many-instance-attributes
    """Object holding information from WLED."""

    architecture: str = field(default="unknown", metadata=field_options(alias="arch"))
    """Name of the platform."""

    arduino_core_version: str = field(
        default="Unknown", metadata=field_options(alias="core")
    )
    """Version of the underlying (Arduino core) SDK."""

    brand: str = "WLED"
    """The producer/vendor of the light. Always WLED for standard installations."""

    build: str = field(default="Unknown", metadata=field_options(alias="vid"))
    """Build ID (YYMMDDB, B = daily build index)."""

    effect_count: int = field(default=0, metadata=field_options(alias="fxcount"))
    """Number of effects included."""

    filesystem: Filesystem = field(metadata=field_options(alias="fs"))
    """Info about the embedded LittleFS filesystem."""

    free_heap: int = field(default=0, metadata=field_options(alias="freeheap"))
    """Bytes of heap memory (RAM) currently available. Problematic if <10k."""

    ip: str = ""  # pylint: disable=invalid-name
    """The IP address of this instance. Empty string if not connected."""

    leds: Leds = field(default_factory=Leds)
    """Contains info about the LED setup."""

    live_ip: str = field(default="Unknown", metadata=field_options(alias="lip"))
    """Realtime data source IP address."""

    live_mode: str = field(default="Unknown", metadata=field_options(alias="lm"))
    """Info about the realtime data source."""

    live: bool = False
    """Realtime data source active via UDP or E1.31."""

    mac_address: str = field(default="", metadata=field_options(alias="mac"))
    """
    The hexadecimal hardware MAC address of the light,
    lowercase and without colons.
    """

    name: str = "WLED Light"
    """Friendly name of the light. Intended for display in lists and titles."""

    palette_count: int = field(default=0, metadata=field_options(alias="palcount"))
    """Number of palettes configured."""

    product: str = "DIY Light"
    """The product name. Always FOSS for standard installations."""

    udp_port: int = field(default=0, metadata=field_options(alias="udpport"))
    """The UDP port for realtime packets and WLED broadcast."""

    uptime: timedelta = timedelta(0)
    """Uptime of the device."""

    version: AwesomeVersion | None = field(
        default=None, metadata=field_options(alias="ver")
    )
    """Version of the WLED software."""

    websocket: int | None = field(default=None, metadata=field_options(alias="ws"))
    """
    Number of currently connected WebSockets clients.
    `None` indicates that WebSockets are unsupported in this build.
    """

    wifi: Wifi | None = None
    """Info about the Wi-Fi connection."""

    @classmethod
    def __post_deserialize__(cls, obj: Info) -> Info:
        """Post deserialize hook for Info object."""
        # If the websocket is disabled in this build, the value will be -1.
        # We want to represent this as None.
        if obj.websocket == -1:
            obj.websocket = None

        # We want the architecture in lower case
        obj.architecture = obj.architecture.lower()

        # We can tweak the architecture name based on the filesystem size.
        if obj.filesystem is not None and obj.architecture == "esp8266":
            if obj.filesystem.total <= 256:
                obj.architecture = "esp01"
            elif obj.filesystem.total <= 512:
                obj.architecture = "esp02"

        return obj


@dataclass(kw_only=True)
class State(BaseModel):
    """Object holding the state of WLED."""

    brightness: int = field(default=1, metadata=field_options(alias="bri"))
    """Brightness of the light.

    If on is false, contains last brightness when light was on (aka brightness
    when on is set to true). Setting bri to 0 is supported but it is
    recommended to use the range 1-255 and use on: false to turn off.

    The state response will never have the value 0 for bri.
    """

    nightlight: Nightlight = field(metadata=field_options(alias="nl"))
    """Nightlight state."""

    on: bool = False
    """The on/off state of the light."""

    playlist_id: int | None = field(default=-1, metadata=field_options(alias="pl"))
    """ID of currently set playlist.."""

    preset_id: int | None = field(default=-1, metadata=field_options(alias="ps"))
    """ID of currently set preset."""

    segments: dict[int, Segment] = field(
        default_factory=dict, metadata=field_options(alias="seg")
    )
    """Segments are individual parts of the LED strip."""

    sync: UDPSync = field(metadata=field_options(alias="udpn"))
    """UDP sync state."""

    transition: int = 0
    """Duration of the crossfade between different colors/brightness levels.

    One unit is 100ms, so a value of 4 results in atransition of 400ms.
    """

    live_data_override: LiveDataOverride = field(metadata=field_options(alias="lor"))
    """Live data override.

    0 is off, 1 is override until live data ends, 2 is override until ESP reboot.
    """

    @classmethod
    def __pre_deserialize__(cls, d: dict[Any, Any]) -> dict[Any, Any]:
        """Pre deserialize hook for State object."""
        # Segments are not indexes, which is suboptimal for the user.
        # We will add the segment ID to the segment data and convert
        # the segments list to an indexed dict.
        d["seg"] = {
            segment_id: segment | {"id": segment_id}
            for segment_id, segment in enumerate(d.get("seg", []))
        }
        return d

    @classmethod
    def __post_deserialize__(cls, obj: State) -> State:
        """Post deserialize hook for State object."""
        # If no playlist is active, the value will be -1. We want to represent
        # this as None.
        if obj.playlist_id == -1:
            obj.playlist_id = None

        # If no preset is active, the value will be -1. We want to represent
        # this as None.
        if obj.preset_id == -1:
            obj.preset_id = None

        return obj


@dataclass(kw_only=True)
class Preset(BaseModel):
    """Object representing a WLED preset."""

    preset_id: int
    """The ID of the preset."""

    name: str = field(default="", metadata=field_options(alias="n"))
    """The name of the preset."""

    quick_label: str | None = field(default=None, metadata=field_options(alias="ql"))
    """The quick label of the preset."""

    on: bool = False
    """The on/off state of the preset."""

    transition: int = 0
    """Duration of the crossfade between different colors/brightness levels.

    One unit is 100ms, so a value of 4 results in atransition of 400ms.
    """

    main_segment_id: int = field(default=0, metadata=field_options(alias="mainseg"))
    """The main segment of the preset."""

    segments: list[Segment] = field(
        default_factory=list, metadata=field_options(alias="seg")
    )
    """Segments are individual parts of the LED strip."""

    @classmethod
    def __pre_deserialize__(cls, d: dict[Any, Any]) -> dict[Any, Any]:
        """Pre deserialize hook for Preset object."""
        # If the segment is a single value, we will convert it to a list.
        if "seg" in d and not isinstance(d["seg"], list):
            d["seg"] = [d["seg"]]

        return d

    @classmethod
    def __post_deserialize__(cls, obj: Preset) -> Preset:
        """Post deserialize hook for Preset object."""
        # If name is empty, we will replace it with the playlist ID.
        if not obj.name:
            obj.name = str(obj.preset_id)
        return obj


@dataclass(frozen=True, kw_only=True)
class PlaylistEntry(BaseModel):
    """Object representing a entry in a WLED playlist."""

    duration: int = field(metadata=field_options(alias="dur"))
    entry_id: int
    preset: int = field(metadata=field_options(alias="ps"))
    transition: int


@dataclass(kw_only=True)
class Playlist(BaseModel):
    """Object representing a WLED playlist."""

    end_preset_id: int | None = field(default=None, metadata=field_options(alias="end"))
    """Single preset ID to apply after the playlist finished.

    Has no effect when an indefinite cycle is set. If not provided,
    the light will stay on the last preset of the playlist.
    """

    entries: list[PlaylistEntry]
    """List of entries in the playlist."""

    name: str = field(default="", metadata=field_options(alias="n"))
    """The name of the playlist."""

    playlist_id: int
    """The ID of the playlist."""

    repeat: int = 0
    """Number of times the playlist should repeat."""

    shuffle: bool = field(default=False, metadata=field_options(alias="r"))
    """Shuffle the playlist entries."""

    @classmethod
    def __pre_deserialize__(cls, d: dict[Any, Any]) -> dict[Any, Any]:
        """Pre deserialize hook for State object."""
        d |= d["playlist"]
        # Duration, presets and transitions values are separate lists stored
        # in the playlist data. We will combine those into a list of
        # dictionaries, which will make it easier to work with the data.
        item_count = len(d.get("ps", []))

        # If the duration is a single value, we will convert it to a list.
        # with the same length as the presets list.
        if not isinstance(d["dur"], list):
            d["dur"] = [d["dur"]] * item_count

        # If the transition value doesn't exists, we will set it to 0.
        if "transitions" not in d:
            d["transitions"] = [0] * item_count
        # If the transition is a single value, we will convert it to a list.
        # with the same length as the presets list.
        elif not isinstance(d["transitions"], list):
            d["transitions"] = [d["transitions"]] * item_count

        # Now we can easily combine the data into a list of dictionaries.
        d["entries"] = [
            {
                "entry_id": entry_id,
                "ps": ps,
                "dur": dur,
                "transition": transition,
            }
            for entry_id, (ps, dur, transition) in enumerate(
                zip(d["ps"], d["dur"], d["transitions"])
            )
        ]

        return d

    @classmethod
    def __post_deserialize__(cls, obj: Playlist) -> Playlist:
        """Post deserialize hook for Playlist object."""
        # If name is empty, we will replace it with the playlist ID.
        if not obj.name:
            obj.name = str(obj.playlist_id)
        return obj


@dataclass(kw_only=True)
class Device(BaseModel):
    """Object holding all information of WLED."""

    info: Info
    state: State

    effects: dict[int, Effect] = field(default_factory=dict)
    palettes: dict[int, Palette] = field(default_factory=dict)
    playlists: dict[int, Playlist] = field(default_factory=dict)
    presets: dict[int, Preset] = field(default_factory=dict)

    @classmethod
    def __pre_deserialize__(cls, d: dict[Any, Any]) -> dict[Any, Any]:
        """Pre deserialize hook for Device object."""
        if (version := d.get("info", {}).get("ver")) and version < MIN_REQUIRED_VERSION:
            msg = (
                f"Unsupported firmware version {version}. "
                f"Minimum required version is {MIN_REQUIRED_VERSION}. "
                f"Please update your WLED device."
            )
            raise WLEDUnsupportedVersionError(msg)

        if _effects := d.get("effects"):
            d["effects"] = {
                effect_id: {"effect_id": effect_id, "name": name}
                for effect_id, name in enumerate(_effects)
            }

        if _palettes := d.get("palettes"):
            d["palettes"] = {
                palette_id: {"palette_id": palette_id, "name": name}
                for palette_id, name in enumerate(_palettes)
            }

        if _presets := d.get("presets"):
            _presets = _presets.copy()
            # The preset data contains both presets and playlists,
            # we split those out, so we can handle those correctly.
            d["presets"] = {
                int(preset_id): preset | {"preset_id": int(preset_id)}
                for preset_id, preset in _presets.items()
                if "playlist" not in preset
                or "ps" not in preset["playlist"]
                or not preset["playlist"]["ps"]
            }
            # Nobody cares about 0.
            d["presets"].pop(0, None)

            d["playlists"] = {
                int(playlist_id): playlist | {"playlist_id": int(playlist_id)}
                for playlist_id, playlist in _presets.items()
                if "playlist" in playlist
                and "ps" in playlist["playlist"]
                and playlist["playlist"]["ps"]
            }
            # Nobody cares about 0.
            d["playlists"].pop(0, None)

        return d

    def update_from_dict(self, data: dict[str, Any]) -> Device:
        """Return Device object from WLED API response.

        Args:
        ----
            data: Update the device object with the data received from a
                WLED device API.

        Returns:
        -------
            The updated Device object.

        """
        if _effects := data.get("effects"):
            self.effects = {
                effect_id: Effect(effect_id=effect_id, name=name)
                for effect_id, name in enumerate(_effects)
            }

        if _palettes := data.get("palettes"):
            self.palettes = {
                palette_id: Palette(palette_id=palette_id, name=name)
                for palette_id, name in enumerate(_palettes)
            }

        if _presets := data.get("presets"):
            # The preset data contains both presets and playlists,
            # we split those out, so we can handle those correctly.
            self.presets = {
                int(preset_id): Preset.from_dict(
                    preset | {"preset_id": int(preset_id)},
                )
                for preset_id, preset in _presets.items()
                if "playlist" not in preset
                or "ps" not in preset["playlist"]
                or not preset["playlist"]["ps"]
            }
            # Nobody cares about 0.
            self.presets.pop(0, None)

            self.playlists = {
                int(playlist_id): Playlist.from_dict(
                    playlist | {"playlist_id": int(playlist_id)}
                )
                for playlist_id, playlist in _presets.items()
                if "playlist" in playlist
                and "ps" in playlist["playlist"]
                and playlist["playlist"]["ps"]
            }
            # Nobody cares about 0.
            self.playlists.pop(0, None)

        if _info := data.get("info"):
            self.info = Info.from_dict(_info)

        if _state := data.get("state"):
            self.state = State.from_dict(_state)

        return self


@dataclass(frozen=True, kw_only=True)
class Releases(BaseModel):
    """Object holding WLED releases information."""

    beta: AwesomeVersion | None
    stable: AwesomeVersion | None
